package structure.integrity;

import static tools.DataStructures.*;

import java.util.*;

import org.antlr.runtime.RecognitionException;

import primitiveType.PredefinedExceptions;
import structure.astNode.Ast;
import structure.astNode.Members;
import structure.astNode.Ast.*;
import structure.integrity.ClassNotWellFormed.*;
import structure.program.ClassTable;
import structure.resource.Resource;
import tools.Assertions;
import tools.DataStructures.Tuple2;
import visitors.*;
public class IntegrityChecker {
  public static final String forbiddenPrefix = "_";
  public static final HashSet<String> primitiveDirectMapType = S(//
      "int",//
      "boolean",//
      "String",//
      "void"//
  );
  public static final HashSet<String> primitiveWrapMapType = S(//
      "class",//
      "name"//
  );
  public static final HashSet<String> classDirectMapType = S(//
      Object.class.getSimpleName(),//
      RuntimeException.class.getSimpleName()//
  );
  public static final HashSet<String> classWrapMapType = S(//
      PredefinedExceptions.CompositionException.class.getSimpleName(),//
      PredefinedExceptions.RestrictException.class.getSimpleName(),//
      PredefinedExceptions.AliasException.class.getSimpleName(),//
      PredefinedExceptions.RedirectException.class.getSimpleName(),//
      PredefinedExceptions.SumException.class.getSimpleName()//
  );
  public static final HashSet<String> predefinedTypesString = concat(primitiveDirectMapType, primitiveWrapMapType);
  public static final HashSet<String> predefinedClassString = concat(classDirectMapType, classWrapMapType);
  public static final HashSet<String> predefinedTypeAndClassString = concat(predefinedTypesString, predefinedClassString);
  public static final ClassName objectClassName = new ClassName("Object");
  public static final ClassName runtimeExceptionClassName = new ClassName(RuntimeException.class.getSimpleName());
  public static final HashSet<ClassName> predefinedThrowables = S(//
      runtimeExceptionClassName,//
      new ClassName(PredefinedExceptions.CompositionException.class.getSimpleName()),//
      new ClassName(PredefinedExceptions.AliasException.class.getSimpleName()),//
      new ClassName(PredefinedExceptions.RedirectException.class.getSimpleName()),//
      new ClassName(PredefinedExceptions.RestrictException.class.getSimpleName()),//
      new ClassName(PredefinedExceptions.SumException.class.getSimpleName())//
  );
  
  public static final HashSet<ClassName> predefinedTypesAndClasses = concat(S(//
      new ClassName("int"),//
      new ClassName("boolean"),//
      new ClassName("String"),//
      new ClassName("void"),//
      new ClassName("class"),//
      new ClassName("name"),//
      objectClassName), predefinedThrowables);
  
  public static void printHintOverError(RecognitionException re) {
    if (re instanceof ClassNotWellFormed) {
      ClassNotWellFormed nwf = (ClassNotWellFormed) re;
      System.err.println(nwf.getClass().getSimpleName() + " around " + nwf.message);
    } else {
      System.err.println("Token '" + ((re.token != null) ? re.token.getText() : "") + "' not expected");
      System.err.println("In line" + re.line + " : " + re.charPositionInLine);
    }
    re.getStackTrace();
  }
  public static void checkForbidden(String s) throws RecognitionException {
    //if (forbiddenId.contains(s)) throw new ForbiddenId(s);
    if (s.startsWith(forbiddenPrefix)) throw new ForbiddenId(s);
  }
  public static void checkDuplication(ArrayList<VariableName> list) throws RecognitionException {
    HashSet<VariableName> set = new HashSet<VariableName>(list);
    if (list.size() != set.size()) {
      list = new ArrayList(list);
      for (VariableName n : set)
        list.remove(n);
      assert !list.isEmpty();
      throw new DuplicatedParameterName(list.iterator().next().value);
    }
  }
  
  public static void checkNoInternalCallInConstructor(Constructor c) throws RecognitionException {
    HashSet<MemberName> hs = new VisitorCollectInternalNames().run(c);
    if (hs.isEmpty()) return;
    throw new UsegeOfIntenalNameIntoConstructor(hs.iterator().next().value);
  }
  //private static void checkNoFreeVariables(BaseClassValue bc) throws RecognitionException {
  //  VisitorCheckNoFreeVariablesZero.run(bc);
  //}
  
  public static BaseClassValue checkBaseClass(ArrayList<ClassName> impl, ArrayList<Ast> members) throws RecognitionException {
    Constructor c = null;
    Members m = new Members();
    for (Ast a : members) {
      if (a instanceof Constructor) {
        if (c != null) throw new MoreThanOneConstructor("");
        c = (Constructor) a;
        continue;
      }
      if (a instanceof Member) {
        Member mm = (Member) a;
        if (m.isContained(Member.class, mm.getName())) throw new DuplicatedMemberName(mm.getName().value);
        m.put(mm.getName(), mm);
        continue;
      }
      throw Assertions.codeNotReachable();
    }
    HashSet<ClassName> implRes = new HashSet<ClassName>(impl);
    if (impl.size() != implRes.size()) {
      impl = new ArrayList<ClassName>(impl);
      for (ClassName cn : implRes)
        impl.remove(cn);
      assert !impl.isEmpty();
      throw new DuplicatedSupertypeName(impl.iterator().next().value);
    }
    if (c == null) c = new Constructor(//
        new ParamsDec(new ArrayList<ClassName>(), new ArrayList<VariableName>()),//
        new ArrayList<MemberName>(), new ArrayList<Expression>());
    Tuple2<Members, Constructor> t = VisitorCompleteIternalNamesZero.run(m, c);
    Members membersRes = t._1;
    Constructor constructorRes = t._2;
    checkNoInternalCallInConstructor(constructorRes);
    BaseClassValue result = new BaseClassValue(implRes, membersRes, constructorRes);
    //checkNoFreeVariables(result);
    VisitorCheckNoVariablesDeclaredTwiceZero.run(result);
    result = VisitorTurnFreeVariablesInClassNameValueZero.run(result);
    checkAllFieldInitializedInConstructor(result);
    return result;
    //return normalizeParameters(result);
  }
  /**Not used now for better visualization of ast.
   * */
  private static BaseClassValue normalizeParameters(BaseClassValue result) {
    Members m = new Members();
    VisitorMemberMakeAbstract v = new VisitorMemberMakeAbstract();
    for (IField f : result.members.values(IField.class))
      m.put(f.getName(), f);
    for (IMethod f : result.members.values(IMethod.class)) {
      int vCount = 0;
      for (VariableName vn : ((AbstractMethod) (f.accept(v))).params.names) {
        VariableName nVn = new VariableName(forbiddenPrefix + (vCount++));
        f = (IMethod) f.accept(new VisitorRenameFromToVariableZero(vn, nVn));
      }
      m.put(f.getName(), f);
    }
    Constructor c = result.constructor;
    int vCount = 0;
    for (VariableName vn : result.constructor.params.names) {
      VariableName nVn = new VariableName("_" + (vCount++));
      c = (Constructor) c.accept(new VisitorRenameFromToVariableZero(vn, nVn));
    }
    return new BaseClassValue(result.impl, m, c);
  }
  public static void checkAllFieldInitializedInConstructor(BaseClassValue bc) throws RecognitionException {
    HashSet<MemberName> inited = new HashSet<MemberName>(bc.constructor.names);
    Set<MemberName> required = bc.members.keySet(Field.class);
    required.removeAll(inited);
    if (!required.isEmpty()) throw new NotInitializedField(required.iterator().next().value);
  }
  public static void checkWellFormedClassTable(ClassTable ct) throws RecognitionException {
    //IResource r = (Resource) null;
    assert ct.values().size() == ct.values(Resource.class).size() : ct.values().size() + " " + ct.values(Resource.class).size();
    for (Resource r : ct.values(Resource.class)) {
      r.expression = VisitorTurnFreeVariablesInClassNameValueZero.run(r.expression);
    }
    VisitorCheckNoVariablesDeclaredTwiceZero.run(ct.mainExpression);
    ct.mainExpression = VisitorTurnFreeVariablesInClassNameValueZero.run(ct.mainExpression);
    //String why = new VisitorNoTypeAsMemberNameOrVariabvleName().checkWhy(ct);
    //if (why != null) throw new ClassNotWellFormed.TypeUsedAsVariableOrMemberName(why);
    Set<ClassName> used = new HashSet<ClassName>();
    for (Resource r : ct.values(Resource.class))
      used.addAll(new VisitorCollectUsedTypesStar().run(r.expression));
    used.addAll(new VisitorCollectUsedTypesStar().run(ct.mainExpression));
    used.removeAll(ct.keySet());
    used.removeAll(IntegrityChecker.predefinedTypesAndClasses);
    if (!used.isEmpty()) throw new ClassNotWellFormed.ClassTableNotClosed(used.iterator().next().toString());
    //all check are performed now, so we can alpharename variables
    for (Resource r : ct.values(Resource.class))
      r.expression = VisitorRenameVariableNoClashMemberNameStar.run(r.expression);
    VisitorCheckMainExpression.run(ct.mainExpression);
  }
  /*
  * no duplicated //DONE
  * par name
  *no duplicated//DONE 
  * member name
  * 
  * no this, int string bool class name, print//DONE
  * in membername, classname, varname
  * 
  * no internal call in field initialization (constructor body)//DONE
  * 
  * all nonAbstract field are initialized into constructor//DONE
  * 
  * no scope hiding (variable new wrt parameters+other variables)
  * no free variable//DONE*2
  * 
  * 
  * no unbounded internal call/field access DONE
  * visitor that turns variable names into field access DONE too
  * 
  * 
  * 
  * no unbound classname (on a class table only)//DONE
  * no type used as membername/variable etc//DONE
  *no duplicated
  * class dec,
  * remember that "int" "boolean" "string" "class" "name" are predefined here.

  * */

}
