package tools;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.*;

/**
 * This is A useful set of function and Data Structures to work better in Java :) This class is developed supposing that all the users will include <code>import static easyJ.DataStructures.*;</code>
 * in the initial part of their java files.
 * 
 * Were a complexity is not specified, it means the complexity is in constant time.
 * 
 * ------
 * 
 * @author Naelh8
 * 
 */
public final class DataStructures {
  
  /**
   * It adds Element e to the Collection c and return the modified Collection Complexity: equal to a call of add on c.
   * 
   * @see java.util.Collection#add
   * @param <T>
   *          the type of the Element
   * @param <C>
   *          the type of the Collection
   * @param c
   *          the Collection
   * @param e
   *          the Element
   * @return the modified Collection
   */
  public final static <T, C extends Collection<? super T>> C accumulate(C c, T e) {
    c.add(e);
    return c;
  }
  /**
   * It removes Element e to the Collection c and return the modified Collection Complexity: equal to a call of remove on c.
   * 
   * @see java.util.Collection#remove
   * @param <T>
   *          the type of the Element
   * @param <C>
   *          the type of the Collection
   * @param c
   *          the Collection
   * @param e
   *          the Element
   * @return the modified Collection
   */
  public final static <T, C extends Collection<? super T>> C decumulate(C c, T e) {
    c.remove(e);
    return c;
  }
  
  /**
   * It returns a new array of the type "array of T". Complexity: O(size)
   * 
   * @param <T>
   *          the type of the array.
   * @param clazz
   *          the Class bringing the type
   * @param size
   *          the size of the resulting array
   * @return new array.
   */
  @SuppressWarnings("unchecked")
  public final static <T> T[] newArray(Class<T> clazz, int size) {
    return (T[]) Array.newInstance(clazz, size);
  }
  /**
   * It wraps in a clean way the Collection.toArray(T[])method. calling toArray(c,T.class) is the same of calling c.toArray(new T[c.size()]); when T is a run time known type. Complexity: O(c.size())
   * 
   * @param <T>
   *          the type resulting array
   * @param clazz
   *          the Class bringing the type
   * @param c
   *          the collection
   * @return new array
   */
  public final static <T> T[] toArray(Class<T> clazz, Collection<? extends T> c) {
    return c.toArray(newArray(clazz, c.size()));
  }
  /**
   * 
   * @param <T>
   * @param <AT>
   * @param clazzA
   *          the type of the base array, ex. int[].class
   * @param clazz
   *          the type of the elements ex. int.class
   * @param c
   * @return
   */
  public final static <T, AT> AT toArrayBase(Class<AT> clazzA, Class<T> clazz, Collection<? extends T> c) {
    Class clazz1 = clazzA.getComponentType();
    assert clazz1.equals(clazz);
    Class clazz2 = wrapperFromBase(clazz1);
    if (clazz2 == null) // wasn't a base
      return (AT) c.toArray(newArray(clazz1, c.size()));
    Object[] temp = c.toArray(newArray(clazz2, c.size()));
    Object result = Array.newInstance(clazz1, temp.length);
    for (int i = 0; i < temp.length; i++)
      Array.set(result, i, temp[i]);
    return (AT) result;
  }
  
  private final static Class wrapperFromBase(Class clazz) {
    if (clazz.equals(int.class)) return Integer.class;
    if (clazz.equals(float.class)) return Float.class;
    if (clazz.equals(double.class)) return Double.class;
    if (clazz.equals(long.class)) return Long.class;
    if (clazz.equals(char.class)) return Character.class;
    if (clazz.equals(byte.class)) return Byte.class;
    if (clazz.equals(boolean.class)) return Boolean.class;
    if (clazz.equals(short.class)) return Short.class;
    return null;
  }
  private final static Class baseFromWrapper(Class clazz) {
    if (clazz.equals(Integer.class)) return int.class;
    if (clazz.equals(Float.class)) return float.class;
    if (clazz.equals(Double.class)) return double.class;
    if (clazz.equals(Long.class)) return long.class;
    if (clazz.equals(Character.class)) return char.class;
    if (clazz.equals(Byte.class)) return byte.class;
    if (clazz.equals(Boolean.class)) return boolean.class;
    if (clazz.equals(Short.class)) return short.class;
    return null;
  }
  
  public static final <T, C extends Collection<T>> C concat(C base, Collection<? extends T>... collections) {
    assert collections.length > 0;
    C result = tools.Cloning.copy(base);
    for (int i = 0; i < collections.length; i++)
      result.addAll(collections[i]);
    return result;
    
  }
  
  /**
   * Base class for all the other tuples. It implements all methods using the conversions toList. The hashCode of a Tuple is the hashCode of the toList representation. Two Tuple are equals iff their
   * toList representation are equals. A Tuple toString works delegating the printing of it's inner elements. Complexity: O(toList().size())
   * 
   * @author Naelh8
   * 
   */
  public abstract static class Tuple implements Cloneable, Serializable {
    
    public abstract ArrayList<Object> toList();
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */@Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || !this.getClass().equals(o.getClass())) return false;
      return this.toList().equals(((Tuple) o).toList());
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#hashCode()
     */@Override
    public int hashCode() {
      return this.toList().hashCode();
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */@Override
    public String toString() {
      String result = "T( ";
      for (Object o : this.toList())
        result += o + ", ";
      result = result.substring(0, result.length() - 2);
      return result + ")";
    }
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#clone()
     */@Override
    public Object clone() {
      try {
        return super.clone();
      } catch (CloneNotSupportedException e) {
        throw Assertions.iDoNotKnowJavaError("clone shound be supported here", e);
      }
    }
  }
  
  /**
   * Tuple0 An empty tuple, useful for generic code.
   * 
   * @author Naelh8
   * 
   */
  public static class Tuple0 extends Tuple {
    protected final static Tuple0 tuple0 = new Tuple0();
    @Override
    public ArrayList<Object> toList() {
      return new ArrayList<Object>();
    }
  }
  /**
   * 
   * @author Naelh8
   * 
   * @param <T1>
   *          the type of the Element _1
   */
  public static class Tuple1<T1> extends Tuple {
    public T1 _1;
    public Tuple1(T1 _1) {
      this._1 = _1;
    }
    /**
     * An ArrayList of Object all and only the elements of the tuple Complexity: O(size of the tuple)
     */
    @Override
    public ArrayList<Object> toList() {
      return AL((Object) _1);
    }
  }
  /**
   * 
   * @author Naelh8
   * 
   * @param <T1>
   *          the type of the Element _1
   * @param <T2>
   *          the type of the Element _2
   */
  public static class Tuple2<T1, T2> extends Tuple {
    public T1 _1;
    public T2 _2;
    public Tuple2(T1 _1, T2 _2) {
      this._1 = _1;
      this._2 = _2;
    }
    /*
     * (non-Javadoc)
     * 
     * @see easyJ.DataStructures.Tuple1#toList()
     */
    @Override
    public ArrayList<Object> toList() {
      return AL((Object) _1, _2);
    }
  }
  
  /**
   * 
   * @author Naelh8
   * 
   * @param <T1>
   *          the type of the Element _1
   * @param <T2>
   *          the type of the Element _2
   * @param <T3>
   *          the type of the Element _3
   */
  public static class Tuple3<T1, T2, T3> extends Tuple {
    public T1 _1;
    public T2 _2;
    public T3 _3;
    public Tuple3(T1 _1, T2 _2, T3 _3) {
      this._1 = _1;
      this._2 = _2;
      this._3 = _3;
    }
    /*
     * (non-Javadoc)
     * 
     * @see easyJ.DataStructures.Tuple2#toList()
     */
    @Override
    public ArrayList<Object> toList() {
      return AL((Object) _1, _2, _3);
    }
  }
  
  /**
   * 
   * @author Naelh8
   * 
   * @param <T1>
   *          the type of the Element _1
   * @param <T2>
   *          the type of the Element _2
   * @param <T3>
   *          the type of the Element _3
   * @param <T4>
   *          the type of the Element _4
   */
  public static class Tuple4<T1, T2, T3, T4> extends Tuple {
    public T1 _1;
    public T2 _2;
    public T3 _3;
    public T4 _4;
    public Tuple4(T1 _1, T2 _2, T3 _3, T4 _4) {
      this._1 = _1;
      this._2 = _2;
      this._3 = _3;
      this._4 = _4;
    }
    /*
     * (non-Javadoc)
     * 
     * @see easyJ.DataStructures.Tuple3#toList()
     */
    @Override
    public ArrayList<Object> toList() {
      return AL((Object) _1, _2, _3, _4);
    }
  }
  
  /**
   * T(_1,_2)returns a Tuple t with t._1==_1 and t._2==_2;
   * 
   * @param <T1>
   *          Type of _1
   * @param <T2>
   *          Type of _2
   * @param _1
   * @param _2
   * @return a new Tuple
   */
  public final static <T1, T2> Tuple2<T1, T2> T(T1 _1, T2 _2) {
    return new Tuple2<T1, T2>(_1, _2);
  }
  
  /**
   * T()returns a Tuple t; Since a Tuple0 is immutable, there exists a single Tuple0 instance that is always returned from this method.
   * 
   * @return a Tuple
   */
  public final static Tuple0 T() {
    return Tuple0.tuple0;
  }
  /**
   * T(_1)returns a Tuple t with t._1==_1;
   * 
   * @param <T1>
   *          Type of _1
   * @param _1
   * @return a new Tuple
   */
  public final static <T1> Tuple1<T1> T(T1 _1) {
    return new Tuple1<T1>(_1);
  }
  
  /**
   * T(_1,_2,_3)returns a Tuple t with t._1==_1, t._2==_2 and t._3==_3;
   * 
   * @param <T1>
   *          Type of _1
   * @param <T2>
   *          Type of _2
   * @param <T3>
   *          Type of _3
   * @param _1
   * @param _2
   * @param _3
   * @return a new Tuple
   */
  public final static <T1, T2, T3> Tuple3<T1, T2, T3> T(T1 _1, T2 _2, T3 _3) {
    return new Tuple3<T1, T2, T3>(_1, _2, _3);
  }
  
  /**
   * T(_1,_2,_3,_4)returns a Tuple t with t._1==_1, t._2==_2, t._3==_3 and t._4==_4;
   * 
   * @param <T1>
   *          Type of _1
   * @param <T2>
   *          Type of _2
   * @param <T3>
   *          Type of _3
   * @param <T4>
   *          Type of _4
   * @param _1
   * @param _2
   * @param _3
   * @param _4
   * @return a new Tuple
   */
  public final static <T1, T2, T3, T4> Tuple4<T1, T2, T3, T4> T(T1 _1, T2 _2, T3 _3, T4 _4) {
    return new Tuple4<T1, T2, T3, T4>(_1, _2, _3, _4);
  }
  
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static int[] AInt(int... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static float[] AFloat(float... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static double[] ADouble(double... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static long[] ALong(long... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static boolean[] ABoolean(boolean... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static byte[] AByte(byte... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static char[] AChar(char... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static short[] AShort(short... ts) {
    return ts;
  }
  /**
   * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param <T>
   *          the type of the Elements
   * @param clazz
   *          the base class the array is able to contain.
   * @param ts
   *          the Elements
   * @return the new array
   */
  public final static <T> T[] A(Class<T> clazz, T... ts) {
    if (ts.getClass().getComponentType().equals(clazz)) return ts;
    // System.out.println("runHere");
    T[] result = newArray(clazz, ts.length);
    for (int i = 0; i < ts.length; i++)
      result[i] = ts[i];
    return result;
  }
  
  /**
   * returns a new ArrayList with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param <T>
   *          the type of the Elements
   * @param ts
   *          the Elements
   * @return the new ArrayList
   */
  public final static <T> ArrayList<T> AL(T... ts) {
    return new ArrayList<T>(Arrays.asList(ts));
  }
  
  /**
   * returns a new Vector with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param <T>
   *          the type of the Elements
   * @param ts
   *          the Elements
   * @return the new Vector
   */
  public final static <T> Vector<T> V(T... ts) {
    return new Vector<T>(Arrays.asList(ts));
  }
  
  /**
   * returns a new HashSet with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
   * 
   * @param <T>
   *          the type of the Elements
   * @param ts
   *          the Elements
   * @return the new HashSet
   */
  public final static <T> HashSet<T> S(T... ts) {
    return new HashSet<T>(Arrays.asList(ts));
  }
  
  /**
   * returns a new HashMap with the pair Key, Elements specified. Note that the Key Element pair have to be provided using a Tuple2 example: M(T("one",1),T("two",2)) will generate a map
   * {"one"=>1,"two"=>2} with keys {"one","two"} and elements {1,2} Complexity: O(ts.length)
   * 
   * @param <K>
   *          the type of the Keys
   * @param <T>
   *          the type of the Elements
   * @param ts
   *          the pairs
   * @return the new HashMap
   */
  public final static <K, T> HashMap<K, T> M(Tuple2<K, T>... ts) {
    HashMap<K, T> result = new HashMap<K, T>(ts.length);
    for (Tuple2<K, T> e : ts)
      result.put(e._1, e._2);
    return result;
  }
  
  public static class WrapperException extends RuntimeException {
    public Object wrappedElement;
    public WrapperException(Object wrappedElement) {
      this.wrappedElement = wrappedElement;
    }
    
  }
  
  /**
   * The class exists only as a container for the functionality, so It have not to be initialized.
   */
  private DataStructures() {}
}
/*
 * returns a new array with the Elements specified. Note that if Elements have a base type, they are automatically boxed. Complexity: O(ts.length)
 * 
 * @param <T> the type of the Elements
 * 
 * @param ts the Elements
 * 
 * @return the new array
 */
// public static <T> T[] A(T...ts){return ts;}
